import torch
import torch.nn as nn


def patch_first_conv(model, in_channels):
    """Change first convolution layer input channels.
    In case:
        in_channels == 1 or in_channels == 2 -> reuse original weights
        in_channels > 3 -> make random kaiming normal initialization
    """

    # get first conv
    for module in model.modules():
        if isinstance(module, nn.Conv2d):
            break

    # change input channels for first conv
    module.in_channels = in_channels
    weight = module.weight.detach()
    reset = False

    if in_channels == 1:
        weight = weight.sum(1, keepdim=True)
    elif in_channels == 2:
        weight = weight[:, :2] * (3.0 / 2.0)
    else:
        reset = True
        weight = torch.Tensor(
            module.out_channels,
            module.in_channels // module.groups,
            *module.kernel_size
        )

    module.weight = nn.parameter.Parameter(weight)
    if reset:
        module.reset_parameters()


def replace_strides_with_dilation(module, dilation_rate):
    """Patch Conv2d modules replacing strides with dilation"""
    for mod in module.modules():
        if isinstance(mod, nn.Conv2d):
            mod.stride = (1, 1)
            mod.dilation = (dilation_rate, dilation_rate)
            kh, kw = mod.kernel_size
            mod.padding = ((kh // 2) * dilation_rate, (kh // 2) * dilation_rate)

            # Kostyl for EfficientNet
            if hasattr(mod, "static_padding"):
                mod.static_padding = nn.Identity()



import numpy as np
import requests
from pathlib import Path
import torch
import torch.nn as nn
from torchvision import transforms
from PIL import Image


class pre_process():
    def load_model(model_path):
        """
        Loads a pre-trained model from a given path.
        Args:
        - model_path: str, the path to the pre-trained model. 
                    If model_path = 'base', then the default pre-trained model will be loaded.
        
        Returns:
        - model: torch.nn.Module, the pre-trained model loaded from the specified path.
        Raises:
        - TypeError: if the input model_path is not 'base'.
        """
        
        if model_path == 'base':
            # Define the path to the default pre-trained model weights
            model_weights_path = "weights.pth"
            path = Path(model_weights_path)

            # If the model weights file does not exist, download it from Dropbox
            if not path.is_file():
                url = "https://www.dropbox.com/s/37rtedwwdslz9w6/all_datasets.pth?dl=1"
                response = requests.get(url)
                open("weights.pth", "wb").write(response.content)
            
            # Load the pre-trained model and set the device to CPU
            model = torch.load(model_weights_path, map_location=torch.device("cpu"))
            # Convert the model to be compatible with multiple GPUs, if available
            model = nn.DataParallel(model.module)
            
            return model
        else:
            raise TypeError("Model not implemented")

    def image_tensor(img):
        """
        Converts a PIL Image or NumPy array to a PyTorch tensor.
        Args:
        - img: PIL.Image or np.ndarray, the image to be converted to a PyTorch tensor.
        Returns:
        - image: torch.Tensor, the converted PyTorch tensor.
        Raises:
        - TypeError: if the input image is not a PIL.Image or np.ndarray.
        """
        
        if type(img) not in [np.ndarray, Image.Image]:
            raise TypeError("Input must be np.ndarray or PIL.Image")

        # Define a PyTorch tensor transformer pipeline
        torch_tensor = transforms.Compose(
            [transforms.Resize((256, 256)), transforms.ToTensor()]
        )

        if type(img) == Image.Image:
            # Convert PIL image to PyTorch tensor
            image = torch_tensor(img)
            image = image.unsqueeze(0)
            print("tensor", type(image))
            return image
        elif type(img) == np.ndarray:
            # Convert NumPy array to RGB PIL image and then to PyTorch tensor
            pil_image = Image.fromarray(img).convert("RGB")
            image = torch_tensor(pil_image)
            image = image.unsqueeze(0)
            print("tensor", type(image))
            return image
        else:
            raise TypeError("Input must be np.ndarray or PIL.Image")




import sys, os
sys.path.insert(0, os.path.abspath('.'))
import numpy as np
import matplotlib.pyplot as plt
import mammopy as mi


class visualization():


    def visualize_breast_dense_segmentation(image, breast_prediction_mask, dense_prediction_mask):
        edges = mi.analysis.canny_edges(breast_prediction_mask)
        #plotttig the results
        combined_sigm, axes = plt.subplots(1,2, figsize = (15,10),squeeze=False)
        axes[0, 0].set_title('Image', fontsize=16)
        axes[0, 1].set_title("Breast and dense tissue segmentation", fontsize=20)
        axes[0, 0].imshow(image, cmap='gray')
        axes[0, 0].set_axis_off()
        axes[0, 1].imshow(image, cmap='gray')
        axes[0, 1].imshow(mi.analysis.mask_to_rgba(edges, color='red'), cmap='gray')
        axes[0, 1].imshow(mi.analysis.mask_to_rgba(dense_prediction_mask, color='green'), cmap='gray', alpha=0.7)
        axes[0, 1].set_axis_off()

        brest_sigm, axes = plt.subplots(1,1, squeeze=False)
        axes[0, 0].imshow(image, cmap='gray')
        axes[0, 0].imshow(mi.analysis.mask_to_rgba(edges, color='red'), cmap='gray')
        axes[0, 0].set_axis_off()

        dens_sigm, axes = plt.subplots(1,1, squeeze=False)
        axes[0, 0].imshow(image, cmap='gray')
        axes[0, 0].imshow(mi.analysis.mask_to_rgba(dense_prediction_mask, color='green'), cmap='gray', alpha=0.7)
        axes[0, 0].set_axis_off()        

        return combined_sigm, brest_sigm, dens_sigm



import sys
import torch
import torch.optim as optim
import torch.nn as nn
import numpy as np
from tqdm import tqdm as tqdm
from .meter import AverageValueMeter

class WeightedLoss(nn.Module):
    def __init__(self, task_num):
        super(WeightedLoss, self).__init__()
        self.task_num = task_num
        self.log_vars = nn.Parameter(torch.zeros((task_num)))

    def forward(self, loss_1, loss_2):
        precision_1 = torch.exp(-self.log_vars[0])
        loss_1 = precision_1 * loss_1 + self.log_vars[0]

        precision_2 = torch.exp(-self.log_vars[1])
        loss_2 = precision_2 * loss_2 + self.log_vars[1]

        return loss_1 + loss_2


class MultiTaskLoss(nn.Module):
    def __init__(self):
        super(MultiTaskLoss, self).__init__()
        #self.tasks = nn.ModuleList(tasks)
        self.sigma_1 = nn.Parameter(torch.zeros(1))
        self.sigma_2 = nn.Parameter(torch.zeros(1))
        #self.mse = nn.MSELoss()

    def forward(self, loss_1, loss_2):
       #l = [self.mse(f(x), y) for y, f in zip(targets, self.tasks)]

       l1 = (torch.Tensor(loss_1) / self.sigma_1**2) + self.sigma_1
       l2 = (torch.Tensor(loss_2) / self.sigma_2 ** 2) + self.sigma_2
       loss = l1 + l2
       return loss

#multitask_loss = MultiTaskLoss()
multitask_loss = WeightedLoss(task_num=2)

class Epoch:

    def __init__(self, model, loss, metrics, stage_name, device='cpu', verbose=True):
        self.model = model
        self.loss = loss
        self.metrics = metrics
        self.stage_name = stage_name
        self.verbose = verbose
        self.device = device

        self._to_device()

    def _to_device(self):
        self.model.to(self.device)
        self.loss.to(self.device)
        for metric in self.metrics:
            metric.to(self.device)

    def _format_logs(self, logs):
        str_logs = ['{} - {:.4}'.format(k, v) for k, v in logs.items()]
        s = ', '.join(str_logs)
        return s

    def batch_update(self, x, y):
        raise NotImplementedError

    def on_epoch_start(self):
        pass

    def run(self, dataloader):

        self.on_epoch_start()

        logs = {}
        loss1_meter = AverageValueMeter()
        loss2_meter = AverageValueMeter()
        loss_meter = AverageValueMeter()
        metrics_meters = {metric.__name__: AverageValueMeter() for metric in self.metrics}
        #metrics_meters2 = {metric.__name__: AverageValueMeter() for metric in self.metrics}

        with tqdm(dataloader, desc=self.stage_name, file=sys.stdout, disable=not (self.verbose)) as iterator:
            for x, y, z in iterator:
                x, y, z = x.to(self.device), y.to(self.device), z.to(self.device)
                loss1, loss2, weighted_loss, y_pred, z_pred = self.batch_update(x, y, z)

                # update loss logs
                loss1_value = loss1.cpu().detach().numpy()
                loss2_value = loss2.cpu().detach().numpy()
                loss_value = weighted_loss.cpu().detach().numpy()

                loss1_meter.add(loss1_value)
                loss2_meter.add(loss2_value)
                loss_meter.add(loss_value)

                loss1_logs = {self.loss.__name__ + str('_breast'):loss1_meter.mean}
                loss2_logs = {self.loss.__name__ + str('_dense'):loss2_meter.mean}
                loss_logs = {self.loss.__name__ + str('_weighted'): loss_meter.mean}

                logs.update(loss1_logs)
                logs.update(loss2_logs)
                logs.update(loss_logs)

                # update metrics logs
                for metric_fn in self.metrics:
                    metric_value1 = metric_fn(y_pred, y).cpu().detach().numpy()
                    metric_value2 = metric_fn(z_pred, z).cpu().detach().numpy()

                    metric_value = (metric_value1 + metric_value2)/2

                    metrics_meters[metric_fn.__name__ ].add(metric_value)
                    #metrics_meters2[metric_fn.__name__ ].add(metric_value2)

                metrics_logs = {k: v.mean for k, v in metrics_meters.items()}
                #metrics_logs2 = {k: v.mean for k, v in metrics_meters2.items()}
                #logs.update(metrics_logs1)
                logs.update(metrics_logs)

                if self.verbose:
                    s = self._format_logs(logs)
                    iterator.set_postfix_str(s)

        return logs


class TrainEpoch(Epoch):

    def __init__(self, model, loss, metrics, optimizer, lr_schedular, device='cuda', verbose=True):
        super().__init__(
            model=model,
            loss=loss,
            metrics=metrics,
            stage_name='train',
            device=device,
            verbose=verbose,
        )
        self.optimizer = optimizer
        self.lr_schedular = lr_schedular

        #self.lr_schedular = optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.5)

    def on_epoch_start(self):
        self.model.train()

    def batch_update(self, x, y, z):
        self.optimizer.zero_grad()
        prediction1, prediction2 = self.model.forward(x)
        loss1 = self.loss(prediction1, y)
        loss2 = self.loss(prediction2, z)

        #weighted_loss = multitask_loss(loss1, loss2)
        #weighted_loss = weighted_loss.to('cuda')
        weighted_loss = 0.5*loss1 + 0.5*loss2

        #loss = 0.5*loss1 + 0.5*loss2
        #loss = awl(loss1, loss2)

        weighted_loss.backward()
        self.optimizer.step()
        #self.lr_schedular.step()
        return loss1, loss2, weighted_loss, prediction1, prediction2


class ValidEpoch(Epoch):

    def __init__(self, model, loss, metrics, device='cuda', verbose=True):
        super().__init__(
            model=model,
            loss=loss,
            metrics=metrics,
            stage_name='valid',
            device=device,
            verbose=verbose,
        )

    def on_epoch_start(self):
        self.model.eval()

    def batch_update(self, x, y, z):
        with torch.no_grad():
            prediction1, prediction2 = self.model.forward(x)
            loss1 = self.loss(prediction1, y)
            loss2 = self.loss(prediction2, z)

            #weighted_loss = multitask_loss(loss1, loss2)
            #weighted_loss = weighted_loss.to('cuda')
            weighted_loss = 0.5*loss1 + 0.5*loss2

            #loss = 0.5*loss1 + 0.5*loss2
            
        return loss1, loss2, weighted_loss, prediction1, prediction2



import torch
import torch.nn.functional as F2

def _take_channels(*xs, ignore_channels=None):
    if ignore_channels is None:
        return xs
    else:
        channels = [channel for channel in range(xs[0].shape[1]) if channel not in ignore_channels]
        xs = [torch.index_select(x, dim=1, index=torch.tensor(channels).to(x.device)) for x in xs]
        return xs


def _threshold(x, threshold=None):
    if threshold is not None:
        return (x > threshold).type(x.dtype)
    else:
        return x


def iou(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate Intersection over Union between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: IoU (Jaccard) score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    intersection = torch.sum(gt * pr)
    union = torch.sum(gt) + torch.sum(pr) - intersection + eps
    return (intersection + eps) / union


jaccard = iou

def tversky_score(pr, gt, alpha=0.3, beta=0.7, eps=1e-7, threshold=None, ignore_channels=None):
    """Calcutate Tversky score between ground truth and prediction"""

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fp = torch.sum(pr) - tp
    fn = torch.sum(gt) - tp

    score = (tp + eps)/(tp + alpha*fp + beta*fn + eps)

    return score



def f_score(pr, gt, beta=1, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate F-score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        beta (float): positive constant
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: F score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fp = torch.sum(pr) - tp
    fn = torch.sum(gt) - tp

    score = ((1 + beta ** 2) * tp + eps) \
            / ((1 + beta ** 2) * tp + beta ** 2 * fn + fp + eps)

    return score


def accuracy(pr, gt, threshold=0.5, ignore_channels=None):
    """Calculate accuracy score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: precision score
    """
    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt == pr, dtype=pr.dtype)
    score = tp / gt.view(-1).shape[0]
    return score


def precision(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate precision score between ground truth and prediction
    Args:
        pr (torch.Tensor): predicted tensor
        gt (torch.Tensor):  ground truth tensor
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: precision score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fp = torch.sum(pr) - tp

    score = (tp + eps) / (tp + fp + eps)

    return score


def recall(pr, gt, eps=1e-7, threshold=None, ignore_channels=None):
    """Calculate Recall between ground truth and prediction
    Args:
        pr (torch.Tensor): A list of predicted elements
        gt (torch.Tensor):  A list of elements that are to be predicted
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: recall score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    tp = torch.sum(gt * pr)
    fn = torch.sum(gt) - tp

    score = (tp + eps) / (tp + fn + eps)

    return score


def binary_cross_entropy_with_logits(pr, gt, threshold=None, ignore_channels=None):
    """Calculate Recall between ground truth and prediction
    Args:
        pr (torch.Tensor): A list of predicted elements
        gt (torch.Tensor):  A list of elements that are to be predicted
        eps (float): epsilon to avoid zero division
        threshold: threshold for outputs binarization
    Returns:
        float: recall score
    """

    pr = _threshold(pr, threshold=threshold)
    pr, gt = _take_channels(pr, gt, ignore_channels=ignore_channels)

    return F2.binary_cross_entropy_with_logits(gt, pr)


""" Each encoder should have following attributes and methods and be inherited from `_base.EncoderMixin`

Attributes:

    _out_channels (list of int): specify number of channels for each encoder feature tensor
    _depth (int): specify number of stages in decoder (in other words number of downsampling operations)
    _in_channels (int): default number of input channels in first Conv2d layer for encoder (usually 3)

Methods:

    forward(self, x: torch.Tensor)
        produce list of features of different spatial resolutions, each feature is a 4D torch.tensor of
        shape NCHW (features should be sorted in descending order according to spatial resolution, starting
        with resolution same as input `x` tensor).

        Input: `x` with shape (1, 3, 64, 64)
        Output: [f0, f1, f2, f3, f4, f5] - features with corresponding shapes
                [(1, 3, 64, 64), (1, 64, 32, 32), (1, 128, 16, 16), (1, 256, 8, 8),
                (1, 512, 4, 4), (1, 1024, 2, 2)] (C - dim may differ)

        also should support number of features according to specified depth, e.g. if depth = 5,
        number of feature tensors = 6 (one with same resolution as input and 5 downsampled),
        depth = 3 -> number of feature tensors = 4 (one with same resolution as input and 3 downsampled).
"""

import torch.nn as nn
from torchvision.models.vgg import VGG
from torchvision.models.vgg import make_layers
from pretrainedmodels.models.torchvision_models import pretrained_settings

from ._base import EncoderMixin

# fmt: off
cfg = {
    'A': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'B': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'D': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'E': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}
# fmt: on


class VGGEncoder(VGG, EncoderMixin):
    def __init__(self, out_channels, config, batch_norm=False, depth=5, **kwargs):
        super().__init__(make_layers(config, batch_norm=batch_norm), **kwargs)
        self._out_channels = out_channels
        self._depth = depth
        self._in_channels = 3
        del self.classifier

    def make_dilated(self, stage_list, dilation_list):
        raise ValueError("'VGG' models do not support dilated mode due to Max Pooling"
                         " operations for downsampling!")

    def get_stages(self):
        stages = []
        stage_modules = []
        for module in self.features:
            if isinstance(module, nn.MaxPool2d):
                stages.append(nn.Sequential(*stage_modules))
                stage_modules = []
            stage_modules.append(module)
        stages.append(nn.Sequential(*stage_modules))
        return stages

    def forward(self, x):
        stages = self.get_stages()

        features = []
        for i in range(self._depth + 1):
            x = stages[i](x)
            features.append(x)

        return features

    def load_state_dict(self, state_dict, **kwargs):
        keys = list(state_dict.keys())
        for k in keys:
            if k.startswith("classifier"):
                state_dict.pop(k)
        super().load_state_dict(state_dict, **kwargs)


vgg_encoders = {
    "vgg11": {
        "encoder": VGGEncoder,
        "pretrained_settings": pretrained_settings["vgg11"],
        "params": {
            "out_channels": (64, 128, 256, 512, 512, 512),
            "config": cfg["A"],
            "batch_norm": False,
        },
    },
    "vgg11_bn": {
        "encoder": VGGEncoder,
        "pretrained_settings": pretrained_settings["vgg11_bn"],
        "params": {
            "out_channels": (64, 128, 256, 512, 512, 512),
            "config": cfg["A"],
            "batch_norm": True,
        },
    },
    "vgg13": {
        "encoder": VGGEncoder,
        "pretrained_settings": pretrained_settings["vgg13"],
        "params": {
            "out_channels": (64, 128, 256, 512, 512, 512),
            "config": cfg["B"],
            "batch_norm": False,
        },
    },
    "vgg13_bn": {
        "encoder": VGGEncoder,
        "pretrained_settings": pretrained_settings["vgg13_bn"],
        "params": {
            "out_channels": (64, 128, 256, 512, 512, 512),
            "config": cfg["B"],
            "batch_norm": True,
        },
    },
    "vgg16": {
        "encoder": VGGEncoder,
        "pretrained_settings": pretrained_settings["vgg16"],
        "params": {
            "out_channels": (64, 128, 256, 512, 512, 512),
            "config": cfg["D"],
            "batch_norm": False,
        },
    },
    "vgg16_bn": {
        "encoder": VGGEncoder,
        "pretrained_settings": pretrained_settings["vgg16_bn"],
        "params": {
            "out_channels": (64, 128, 256, 512, 512, 512),
            "config": cfg["D"],
            "batch_norm": True,
        },
    },
    "vgg19": {
        "encoder": VGGEncoder,
        "pretrained_settings": pretrained_settings["vgg19"],
        "params": {
            "out_channels": (64, 128, 256, 512, 512, 512),
            "config": cfg["E"],
            "batch_norm": False,
        },
    },
    "vgg19_bn": {
        "encoder": VGGEncoder,
        "pretrained_settings": pretrained_settings["vgg19_bn"],
        "params": {
            "out_channels": (64, 128, 256, 512, 512, 512),
            "config": cfg["E"],
            "batch_norm": True,
        },
    },
}



import functools
import torch.utils.model_zoo as model_zoo

from .resnet import resnet_encoders
from .dpn import dpn_encoders
from .vgg import vgg_encoders
from .senet import senet_encoders
from .densenet import densenet_encoders
from .inceptionresnetv2 import inceptionresnetv2_encoders
from .inceptionv4 import inceptionv4_encoders
from .efficientnet import efficient_net_encoders
from .mobilenet import mobilenet_encoders
from .xception import xception_encoders
from .timm_efficientnet import timm_efficientnet_encoders

from ._preprocessing import preprocess_input

encoders = {}
encoders.update(resnet_encoders)
encoders.update(dpn_encoders)
encoders.update(vgg_encoders)
encoders.update(senet_encoders)
encoders.update(densenet_encoders)
encoders.update(inceptionresnetv2_encoders)
encoders.update(inceptionv4_encoders)
encoders.update(efficient_net_encoders)
encoders.update(mobilenet_encoders)
encoders.update(xception_encoders)
encoders.update(timm_efficientnet_encoders)


def get_encoder(name, in_channels=3, depth=5, weights=None):
    Encoder = encoders[name]["encoder"]
    params = encoders[name]["params"]
    params.update(depth=depth)
    encoder = Encoder(**params)

    if weights is not None:
        settings = encoders[name]["pretrained_settings"][weights]
        encoder.load_state_dict(model_zoo.load_url(settings["url"]))

    encoder.set_in_channels(in_channels)

    return encoder


def get_encoder_names():
    return list(encoders.keys())


def get_preprocessing_params(encoder_name, pretrained="imagenet"):
    settings = encoders[encoder_name]["pretrained_settings"]

    if pretrained not in settings.keys():
        raise ValueError("Avaliable pretrained options {}".format(settings.keys()))

    formatted_settings = {}
    formatted_settings["input_space"] = settings[pretrained].get("input_space")
    formatted_settings["input_range"] = settings[pretrained].get("input_range")
    formatted_settings["mean"] = settings[pretrained].get("mean")
    formatted_settings["std"] = settings[pretrained].get("std")
    return formatted_settings


def get_preprocessing_fn(encoder_name, pretrained="imagenet"):
    params = get_preprocessing_params(encoder_name, pretrained=pretrained)
    return functools.partial(preprocess_input, **params)



import torch
import torch.nn as nn
from typing import List
from collections import OrderedDict

from . import _utils as utils


class EncoderMixin:
    """Add encoder functionality such as:
        - output channels specification of feature tensors (produced by encoder)
        - patching first convolution for arbitrary input channels
    """

    @property
    def out_channels(self):
        """Return channels dimensions for each tensor of forward output of encoder"""
        return self._out_channels[: self._depth + 1]

    def set_in_channels(self, in_channels):
        """Change first convolution chennels"""
        if in_channels == 3:
            return

        self._in_channels = in_channels
        if self._out_channels[0] == 3:
            self._out_channels = tuple([in_channels] + list(self._out_channels)[1:])

        utils.patch_first_conv(model=self, in_channels=in_channels)

    def get_stages(self):
        """Method should be overridden in encoder"""
        raise NotImplementedError

    def make_dilated(self, stage_list, dilation_list):
        stages = self.get_stages()
        for stage_indx, dilation_rate in zip(stage_list, dilation_list):
            utils.replace_strides_with_dilation(
                module=stages[stage_indx],
                dilation_rate=dilation_rate,
            )
